import * as http from 'http';
import * as https from 'https';
import { constants } from 'http2';
import { pluralize } from 'inflection';
import {BaseSchema, ObjectSchema, Output} from 'valibot';
import { SerializerPair } from './serializers';
import {
	handleCreateItem, handleDeleteItem, handleEmplaceItem,
	handleGetCollection,
	handleGetItem,
	handleGetRoot,
	handleHasMethodAndUrl, handlePatchItem,
} from './handlers';
import Negotiator from 'negotiator';
import {getMethod, getUrl} from './utils';
import {EncodingPair} from './encodings';
import * as en from './languages/en';
import * as utf8 from './encodings/utf-8';
import * as applicationJson from './serializers/application/json';

// TODO separate frontend and backend factory methods

export interface DataSource<T = object> {
	initialize(): Promise<unknown>;
	getTotalCount?(): Promise<number>;
	getMultiple(): Promise<T[]>;
	getSingle(id: string): Promise<T | null>;
	create(data: T): Promise<T>;
	delete(id: string): Promise<unknown>;
	emplace(id: string, data: T): Promise<[T, boolean]>;
	patch(id: string, data: Partial<T>): Promise<T | null>;
}

export interface ApplicationParams {
	name: string;
	dataSource?: (resource: Resource) => DataSource;
}

interface ResourceState<T extends BaseSchema> {
	idAttr: string;
	itemName: string;
	collectionName: string;
	routeName: string;
	idConfig: ResourceIdConfig<T>;
	fullTextAttrs: Set<string>;
	canCreate: boolean;
	canFetchCollection: boolean;
	canFetchItem: boolean;
	canPatch: boolean;
	canEmplace: boolean;
	canDelete: boolean;
}

export interface Resource<T extends BaseSchema = any, IdSchema extends BaseSchema = any> {
	newId(dataSource: DataSource): string | number | unknown;
	schema: T;
	state: ResourceState<IdSchema>;
	id(newIdAttr: string, params: ResourceIdConfig<IdSchema>): this;
	fullText(fullTextAttr: string): this;
	name(n: string): this;
	collection(n: string): this;
	route(n: string): this;
	canFetchCollection(b?: boolean): this;
	canFetchItem(b?: boolean): this;
	canCreate(b?: boolean): this;
	canPatch(b?: boolean): this;
	canEmplace(b?: boolean): this;
	canDelete(b?: boolean): this;
}

export interface ResourceWithDataSource<T extends BaseSchema = any> extends Resource<T> {
	dataSource: DataSource;
}

interface GenerationStrategy {
	(dataSource: DataSource, ...args: unknown[]): Promise<string | number | unknown>;
}

interface ResourceIdConfig<T extends BaseSchema> {
	generationStrategy: GenerationStrategy;
	serialize: (id: unknown) => string;
	deserialize: (id: string) => Output<T>;
	schema: T;
}

const getAllowedMiddlewares = (resource: Resource, mainResourceId: string) => {
	const middlewares = [] as [string, Middleware][];
	if (mainResourceId === '') {
		if (resource.state.canFetchCollection) {
			middlewares.push(['GET', handleGetCollection]);
		}
		if (resource.state.canCreate) {
			middlewares.push(['POST', handleCreateItem]);
		}
		return middlewares;
	}

	if (resource.state.canFetchItem) {
		middlewares.push(['GET', handleGetItem]);
	}
	if (resource.state.canEmplace) {
		middlewares.push(['PUT', handleEmplaceItem]);
	}
	if (resource.state.canPatch) {
		middlewares.push(['PATCH', handlePatchItem]);
	}
	if (resource.state.canDelete) {
		middlewares.push(['DELETE', handleDeleteItem]);
	}

	return middlewares;
};

export const resource = <T extends BaseSchema, Id extends BaseSchema = any>(schema: T): Resource<T, Id> => {
	const resourceState = {
		fullTextAttrs: new Set<string>(),
		canCreate: false,
		canFetchCollection: false,
		canFetchItem: false,
		canPatch: false,
		canEmplace: false,
		canDelete: false,
	} as Partial<ResourceState<Id>>;

	return {
		get state(): ResourceState<Id> {
			return Object.freeze({
				...resourceState
			}) as unknown as ResourceState<Id>;
		},
		canFetchCollection(b = true) {
			resourceState.canFetchCollection = b;
			return this;
		},
		canFetchItem(b = true) {
			resourceState.canFetchItem = b;
			return this;
		},
		canCreate(b = true) {
			resourceState.canCreate = b;
			return this;
		},
		canPatch(b = true) {
			resourceState.canPatch = b;
			return this;
		},
		canEmplace(b = true) {
			resourceState.canEmplace = b;
			return this;
		},
		canDelete(b = true) {
			resourceState.canDelete = b;
			return this;
		},
		id(newIdAttr: string, params: ResourceIdConfig<Id>) {
			resourceState.idAttr = newIdAttr;
			resourceState.idConfig = params;
			return this;
		},
		newId(dataSource: DataSource) {
			return resourceState?.idConfig?.generationStrategy?.(dataSource);
		},
		fullText(fullTextAttr: string) {
			if (
				schema.type === 'object'
				&& (
					schema as unknown as ObjectSchema<
						Record<string, BaseSchema>,
						undefined,
						Record<string, string>
					>
				)
					.entries[fullTextAttr]?.type === 'string'
			) {
				resourceState.fullTextAttrs?.add(fullTextAttr);
				return this;
			}

			throw new Error(`Could not set attribute ${fullTextAttr} as fulltext.`);
		},
		name(n: string) {
			resourceState.itemName = n;
			resourceState.collectionName = resourceState.collectionName ?? pluralize(n).toLowerCase();
			resourceState.routeName = resourceState.routeName ?? resourceState.collectionName;
			return this;
		},
		collection(n: string) {
			resourceState.collectionName = n;
			resourceState.routeName = resourceState.routeName ?? n;
			return this;
		},
		route(n: string) {
			resourceState.routeName = n;
			return this;
		},
		get idAttr() {
			return resourceState.idAttr;
		},
		get collectionName() {
			return resourceState.collectionName;
		},
		get itemName() {
			return resourceState.itemName;
		},
		get routeName() {
			return resourceState.routeName;
		},
		get schema() {
			return schema;
		}
	} as Resource;
};

interface CreateServerParams {
	baseUrl?: string;
	host?: string;
	cert?: string;
	key?: string;
	requestTimeout?: number;
}

type RequestListenerWithReturn<
	P extends unknown = unknown,
	Q extends typeof http.IncomingMessage = typeof http.IncomingMessage,
	R extends typeof http.ServerResponse = typeof http.ServerResponse
> = (
	...args: Parameters<http.RequestListener<Q, R>>
) => P;

interface HandlerState {
	handled: boolean;
}

export interface ApplicationState {
	resources: Set<ResourceWithDataSource>;
	languages: Map<string, MessageCollection>;
	serializers: Map<string, SerializerPair>;
	encodings: Map<string, EncodingPair>;
}

type MessageBody = string | string[] | (string | string[])[];

export interface MessageCollection {
	statusMessages: {
		unableToInitializeResourceDataSource(resource: Resource): string;
		unableToFetchResourceCollection(resource: Resource): string;
		unableToFetchResource(resource: Resource): string;
		languageNotAcceptable(): string;
		encodingNotAcceptable(): string;
		mediaTypeNotAcceptable(): string;
		methodNotAllowed(): string;
		urlNotFound(): string;
		badRequest(): string;
		ok(): string;
		resourceCollectionFetched(resource: Resource): string;
		resourceFetched(resource: Resource): string;
		resourceNotFound(resource: Resource): string;
		deleteNonExistingResource(resource: Resource): string;
		unableToSerializeResponse(): string;
		unableToEncodeResponse(): string;
		unableToDeleteResource(resource: Resource): string;
		resourceDeleted(resource: Resource): string;
		unableToDeserializeRequest(): string;
		patchNonExistingResource(resource: Resource): string;
		unableToPatchResource(resource: Resource): string;
		invalidResourcePatch(resource: Resource): string;
		invalidResource(resource: Resource): string;
		resourcePatched(resource: Resource): string;
		resourceCreated(resource: Resource): string;
		resourceReplaced(resource: Resource): string;
	},
	bodies: {
		languageNotAcceptable(): MessageBody;
		encodingNotAcceptable(): MessageBody;
		mediaTypeNotAcceptable(): MessageBody;
	}
}

export interface BackendState {
	fallback: {
		language: string;
		encoding: string;
		serializer: string;
	}
	errorHeaders: {
		language?: string;
		encoding?: string;
		serializer?: string;
	}
	showTotalItemCountOnGetCollection: boolean;
	throws404OnDeletingNotFound: boolean;
	checksSerializersOnDelete: boolean;
	showTotalItemCountOnCreateItem: boolean;
}

interface MiddlewareArgs {
	handlerState: HandlerState;
	backendState: BackendState;
	appState: ApplicationState;
	appParams: ApplicationParams;
	serverParams: CreateServerParams;
	responseBodyLanguage: [string, MessageCollection];
	responseBodyEncoding: [string, EncodingPair];
	responseBodyMediaType: [string, SerializerPair];
	errorResponseBodyLanguage: [string, MessageCollection];
	errorResponseBodyEncoding: [string, EncodingPair];
	errorResponseBodyMediaType: [string, SerializerPair];
	resource: ResourceWithDataSource;
	resourceId: string;
	query: URLSearchParams;
}

export interface Middleware {
	(args: MiddlewareArgs): RequestListenerWithReturn<HandlerState | Promise<HandlerState>>
}

export interface Backend {
	showTotalItemCountOnGetCollection(b?: boolean): this;
	showTotalItemCountOnCreateItem(b?: boolean): this;
	checksSerializersOnDelete(b?: boolean): this;
	throws404OnDeletingNotFound(b?: boolean): this;
	createServer(serverParams?: CreateServerParams): http.Server | https.Server;
}

export interface Client {
	setLanguage(languageCode: string): this;
	setEncoding(encoding: string): this;
	setContentType(contentType: string): this;
}

export interface Application {
	contentType(mimeTypePrefix: string, serializerPair: SerializerPair): this;
	language(languageCode: string, messageCollection: MessageCollection): this;
	encoding(encoding: string, encodingPair: EncodingPair): this;
	resource(resRaw: Partial<Resource>): this;
	createBackend(): Backend;
	createClient(): Client;
}

export const application = (appParams: ApplicationParams): Application => {
	const appState: ApplicationState = {
		resources: new Set<ResourceWithDataSource>(),
		languages: new Map<string, MessageCollection>(),
		serializers: new Map<string, SerializerPair>(),
		encodings: new Map<string, EncodingPair>(),
	};

	appState.languages.set(en.name, en.messages);
	appState.encodings.set(utf8.name, utf8);
	appState.serializers.set(applicationJson.name, applicationJson);

	return {
		contentType(mimeTypePrefix: string, serializerPair: SerializerPair) {
			appState.serializers.set(mimeTypePrefix, serializerPair);
			return this;
		},
		encoding(encoding: string, encodingPair: EncodingPair) {
			appState.encodings.set(encoding, encodingPair);
			return this;
		},
		language(languageCode: string, messageCollection: MessageCollection) {
			appState.languages.set(languageCode, messageCollection);
			return this;
		},
		resource(resRaw: Partial<Resource>) {
			const res = resRaw as Partial<ResourceWithDataSource>;
			res.dataSource = res.dataSource ?? appParams.dataSource?.(res as Resource);
			if (typeof res.dataSource === 'undefined') {
				throw new Error(`Resource ${res.state!.itemName} must have a data source.`);
			}
			appState.resources.add(res as ResourceWithDataSource);
			return this;
		},
		createClient(): Client {
			const clientState = {
				contentType: applicationJson.name,
				encoding: utf8.name,
				language: en.name
			};

			return {
				setContentType(contentType: string) {
					clientState.contentType = contentType;
					return this;
				},
				setEncoding(encoding: string) {
					clientState.encoding = encoding;
					return this;
				},
				setLanguage(languageCode: string) {
					clientState.language = languageCode;
					return this;
				}
			} satisfies Client;
		},
		createBackend(): Backend {
			const backendState: BackendState = {
				fallback: {
					language: en.name,
					encoding: utf8.name,
					serializer: applicationJson.name
				},
				errorHeaders: {
					// undefined follows user accept headers strictly
					//
					language: undefined,
					encoding: undefined,
					serializer: undefined,
				},
				showTotalItemCountOnGetCollection: false,
				showTotalItemCountOnCreateItem: false,
				throws404OnDeletingNotFound: false,
				checksSerializersOnDelete: false,

			};

			return {
				showTotalItemCountOnGetCollection(b = true) {
					backendState.showTotalItemCountOnGetCollection = b;
					return this;
				},
				showTotalItemCountOnCreateItem(b = true) {
					backendState.showTotalItemCountOnCreateItem = b;
					return this;
				},
				throws404OnDeletingNotFound(b = true) {
					backendState.throws404OnDeletingNotFound = b;
					return this;
				},
				checksSerializersOnDelete(b = true) {
					backendState.checksSerializersOnDelete = b;
					return this;
				},
				createServer(serverParams = {} as CreateServerParams) {
					const server = 'key' in serverParams && 'cert' in serverParams
						? https.createServer({
							key: serverParams.key,
							cert: serverParams.cert,
							requestTimeout: serverParams.requestTimeout
						})
						: http.createServer({
							requestTimeout: serverParams.requestTimeout
						});

					server.on('request', async (req, res) => {
						const method = getMethod(req);
						const baseUrl = serverParams.baseUrl ?? '';
						const { url, query } = getUrl(req, baseUrl);

						const negotiator = new Negotiator(req);
						const languageCandidate = negotiator.language(Array.from(appState.languages.keys())) ?? backendState.fallback.language;
						const encodingCandidate = negotiator.encoding(Array.from(appState.encodings.keys())) ?? backendState.fallback.encoding;
						const contentTypeCandidate = negotiator.mediaType(Array.from(appState.serializers.keys())) ?? backendState.fallback.serializer;

						const availableLanguages = Array.from(appState.languages.entries());
						const fallbackMessageCollection = en.messages as MessageCollection;
						const fallbackSerializerPair = applicationJson as SerializerPair;
						const fallbackEncoding = utf8 as EncodingPair;

						const errorLanguageCode = backendState.errorHeaders.language ?? backendState.fallback.language;
						const errorMessageCollection = appState.languages.get(errorLanguageCode) ?? fallbackMessageCollection;

						const errorContentType = backendState.errorHeaders.serializer ?? backendState.fallback.serializer;
						const errorSerializerPair = appState.serializers.get(errorContentType) ?? fallbackSerializerPair;

						const errorEncodingKey = backendState.errorHeaders.encoding ?? backendState.fallback.encoding;
						const errorEncoding = appState.encodings.get(errorEncodingKey) ?? fallbackEncoding;

						const [currentLanguageCode, currentLanguageMessages] = availableLanguages.find(([code]) => code === languageCandidate) ?? [];
						if (typeof currentLanguageCode === 'undefined' || typeof currentLanguageMessages === 'undefined') {
							const data = errorMessageCollection.bodies.languageNotAcceptable();
							const responseRaw = errorSerializerPair.serialize(data);
							const response = errorEncoding.encode(responseRaw);
							res.writeHead(constants.HTTP_STATUS_NOT_ACCEPTABLE, {
								'Content-Language': errorLanguageCode,
								'Content-Type': errorContentType,
								'Content-Encoding': errorEncodingKey,
							});
							res.statusMessage = errorMessageCollection.statusMessages.languageNotAcceptable();
							res.end(response);
							return;
						}

						const availableMediaTypes = Array.from(appState.serializers.entries());
						const [currentContentTypeMimeType, responseMediaTypeEntry] = availableMediaTypes.find(([key]) => key === contentTypeCandidate) ?? [];
						if (typeof currentContentTypeMimeType === 'undefined' || typeof responseMediaTypeEntry === 'undefined') {
							const data = errorMessageCollection.bodies.languageNotAcceptable();
							const responseRaw = errorSerializerPair.serialize(data);
							const response = errorEncoding.encode(responseRaw);
							res.writeHead(constants.HTTP_STATUS_NOT_ACCEPTABLE, {
								'Content-Language': errorLanguageCode,
								'Content-Type': errorContentType,
								'Content-Encoding': errorEncodingKey,
							});
							res.statusMessage = errorMessageCollection.statusMessages.mediaTypeNotAcceptable();
							res.end(response);
							return;
						}

						const availableEncodings = Array.from(appState.encodings.entries());
						const [currentEncoding, responseBodyEncodingEntry] = availableEncodings.find(([key]) => key === encodingCandidate) ?? [];
						if (typeof currentEncoding === 'undefined' || typeof responseBodyEncodingEntry === 'undefined') {
							const data = errorMessageCollection.bodies.languageNotAcceptable();
							const responseRaw = errorSerializerPair.serialize(data);
							const response = errorEncoding.encode(responseRaw);
							res.writeHead(constants.HTTP_STATUS_NOT_ACCEPTABLE, {
								'Content-Language': errorLanguageCode,
								'Content-Type': errorContentType,
								'Content-Encoding': errorEncodingKey,
							});
							res.statusMessage = errorMessageCollection.statusMessages.encodingNotAcceptable();
							res.end(response);
							return;
						}

						const middlewareArgs: Omit<MiddlewareArgs, 'resource' | 'resourceId'> = {
							handlerState: {
								handled: false
							},
							appState,
							appParams,
							backendState,
							serverParams,
							query,
							responseBodyEncoding: [currentEncoding, responseBodyEncodingEntry],
							responseBodyMediaType: [currentContentTypeMimeType, responseMediaTypeEntry],
							responseBodyLanguage: [currentLanguageCode, currentLanguageMessages],
							errorResponseBodyMediaType: [errorContentType, errorSerializerPair],
							errorResponseBodyEncoding: [errorEncodingKey, errorEncoding],
							errorResponseBodyLanguage: [errorLanguageCode, errorMessageCollection],
						};

						const methodAndUrl = await handleHasMethodAndUrl(middlewareArgs as MiddlewareArgs)(req, res);
						if (methodAndUrl.handled) {
							return;
						}

						if (url === '/') {
							const middlewareState = await handleGetRoot(middlewareArgs as MiddlewareArgs)(req, res);
							if (middlewareState.handled) {
								return;
							}

							res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
								Allow: 'HEAD, GET'
							});
							res.statusMessage = errorMessageCollection.statusMessages.methodNotAllowed();
							res.end();
							return;
						}

						const [, resourceRouteName, resourceId = ''] = url.split('/');
						const resource = Array.from(appState.resources).find((r) => r.state!.routeName === resourceRouteName);
						if (typeof resource === 'undefined') {
							res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
							res.statusMessage = errorMessageCollection.statusMessages.urlNotFound();
							res.end();
							return;
						}

						const middlewares = getAllowedMiddlewares(resource, resourceId);
						const middlewareState = await middlewares
							.reduce(
								async (currentHandlerStatePromise, [middlewareMethod, middleware]) => {
									const currentHandlerState = await currentHandlerStatePromise;
									if (method !== middlewareMethod) {
										return currentHandlerState;
									}

									if (currentHandlerState.handled) {
										return currentHandlerState;
									}

									return middleware({
										...middlewareArgs,
										handlerState: currentHandlerState,
										resource,
										resourceId: resourceId,
									})(req, res);
								},
								Promise.resolve<HandlerState>({
									handled: false
								})
							);

						if (middlewareState.handled) {
							return;
						}

						if (middlewares.length > 0) {
							res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
								Allow: middlewares.map((m) => m[0]).join(', ')
							});
							res.statusMessage = errorMessageCollection.statusMessages.methodNotAllowed();
							res.end();
							return;
						}

						res.statusCode = constants.HTTP_STATUS_NOT_FOUND;
						res.statusMessage = errorMessageCollection.statusMessages.urlNotFound();
						res.end();
						return;
					});

					return server;
				}
			} satisfies Backend;
		},
	};
};
