import { constants } from 'http2';
import * as v from 'valibot';
import {Middleware} from './core';
import {getBody, getDeserializerObjects} from './utils';
import {IncomingMessage, ServerResponse} from 'http';

export const handleHasMethodAndUrl: Middleware = ({
	errorResponseBodyLanguage: [errorLanguageCode, errorMessageCollection],
}) => (req: IncomingMessage, res: ServerResponse) => {
	if (!req.method) {
		res.writeHead(constants.HTTP_STATUS_METHOD_NOT_ALLOWED, {
			'Allow': 'HEAD, GET, POST, PUT, PATCH, DELETE',
			'Content-Language': errorLanguageCode,

		});
		res.statusMessage = errorMessageCollection.statusMessages.methodNotAllowed();
		res.end();
		return {
			handled: true
		};
	}

	if (!req.url) {
		res.statusCode = constants.HTTP_STATUS_BAD_REQUEST;
		res.statusMessage = errorMessageCollection.statusMessages.badRequest();
		res.end();
		return {
			handled: true
		};
	}

	return {
		handled: false
	};
};

export const handleGetRoot: Middleware = ({
	appState,
	appParams,
	serverParams,
	responseBodyMediaType: [responseBodyMediaType, responseBodySerializerPair],
	responseBodyLanguage: [languageCode, responseBodyMessageCollection],
	responseBodyEncoding: [encodingKey, encoding],
	errorResponseBodyLanguage: [errorLanguageCode, errorMessageCollection],
}) => (_req: IncomingMessage, res: ServerResponse) => {
	const singleResDatum = {
		name: appParams.name
	};

	let serialized;
	try {
		serialized = responseBodySerializerPair.serialize(singleResDatum);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let theFormatted;
	try {
		theFormatted = encoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	const theHeaders: Record<string, string> = {
		'Content-Type': responseBodyMediaType,
		'Content-Language': languageCode,
		'Content-Encoding': encodingKey,
	};

	const registeredResources = Array.from(appState.resources);
	const availableResources = registeredResources.filter((r) => (
		r.canFetchCollection
		|| r.canCreate
	));

	if (availableResources.length > 0) {
		// we are using custom headers for links because the standard Link header
		// is referring to the document metadata (e.g. author, next page, etc)
		// https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Link
		theHeaders['X-Resource-Link'] = availableResources
			.map((r) =>
				`<${serverParams.baseUrl}/${r.state.routeName}>; name="${r.state.collectionName}"`,
			)
			.join(', ');
	}
	res.writeHead(constants.HTTP_STATUS_OK, theHeaders);
	res.statusMessage = responseBodyMessageCollection.statusMessages.ok();
	res.end(theFormatted);
	return {
		handled: true
	};
};

export const handleGetCollection: Middleware = ({
	resource,
	responseBodyMediaType: [responseBodyMediaType, responseBodySerializerPair],
	responseBodyLanguage: [languageCode, responseBodyMessageCollection],
	responseBodyEncoding: [encodingKey, encoding],
	errorResponseBodyLanguage: [errorLanguageCode, errorMessageCollection],
	backendState,
}) => async (_req: IncomingMessage, res: ServerResponse) => {
	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let resData: Object[];
	let totalItemCount: number | undefined;
	try {
		// TODO querying mechanism
		resData = await resource.dataSource.getMultiple(); // TODO paginated responses per resource
		if (backendState.showTotalItemCountOnGetCollection && typeof resource.dataSource.getTotalCount === 'function') {
			totalItemCount = await resource.dataSource.getTotalCount();
		}
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToFetchResourceCollection(resource);
		res.end();
		return {
			handled: true
		};
	}

	let serialized;
	try {
		serialized = responseBodySerializerPair.serialize(resData);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let theFormatted;
	try {
		theFormatted = encoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	const headers: Record<string, string> = {
		'Content-Type': responseBodyMediaType,
		'Content-Language': languageCode,
		'Content-Encoding': encodingKey,
	};

	if (typeof totalItemCount !== 'undefined') {
		headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
	}

	res.writeHead(constants.HTTP_STATUS_OK, headers);
	res.statusMessage = responseBodyMessageCollection.statusMessages.resourceCollectionFetched(resource);
	res.end(theFormatted);
	return {
		handled: true
	};
};

export const handleGetItem: Middleware = ({
	resourceId,
	resource,
	responseBodyMediaType: [responseBodyMediaType, responseBodySerializerPair],
	responseBodyLanguage: [languageCode, responseBodyMessageCollection],
	responseBodyEncoding: [encodingKey, encoding],
	errorResponseBodyLanguage: [errorLanguageCode, errorMessageCollection],
}) => async (_req: IncomingMessage, res: ServerResponse) => {
	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let singleResDatum: Object | null = null;
	try {
		singleResDatum = await resource.dataSource.getSingle(resourceId);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToFetchResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let serialized: string | null;
	try {
		serialized = singleResDatum === null ? null : responseBodySerializerPair.serialize(singleResDatum);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let theFormatted;
	try {
		theFormatted = serialized === null ? null : encoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	if (theFormatted) {
		res.writeHead(constants.HTTP_STATUS_OK, {
			'Content-Type': responseBodyMediaType,
			'Content-Language': languageCode,
			'Content-Encoding': encodingKey,
		});
		res.statusMessage = responseBodyMessageCollection.statusMessages.resourceFetched(resource)
		res.end(theFormatted);
		return {
			handled: true
		};
	}

	res.writeHead(constants.HTTP_STATUS_NOT_FOUND, {
		'Content-Language': errorLanguageCode,
	});
	res.statusMessage = errorMessageCollection.statusMessages.resourceNotFound(resource);
	res.end();
	return {
		handled: true
	};
};


export const handleDeleteItem: Middleware = ({
	resource,
	resourceId,
	responseBodyLanguage: [languageCode, responseBodyMessageCollection],
	errorResponseBodyLanguage: [errorLanguageCode, errorMessageCollection],
	backendState,
}) => async (_req: IncomingMessage, res: ServerResponse) => {
	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let response;
	try {
		response = await resource.dataSource.delete(resourceId);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToDeleteResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	const throwOnNotFound = !response && backendState.throws404OnDeletingNotFound;

	res.writeHead(
		throwOnNotFound
			? constants.HTTP_STATUS_NOT_FOUND
			: constants.HTTP_STATUS_NO_CONTENT,

		throwOnNotFound
			? {
				'Content-Language': errorLanguageCode,
				// TODO provide more details
			}
			: {
				'Content-Language': languageCode,
			}
	);
	res.statusMessage = (
		throwOnNotFound
			? errorMessageCollection.statusMessages.deleteNonExistingResource(resource)
			: responseBodyMessageCollection.statusMessages.resourceDeleted(resource)
	);

	if (throwOnNotFound) {
		// TODO provide error message
		res.end();
		return {
			handled: true
		};
	}

	res.end();
	return {
		handled: true
	};
};

export const handlePatchItem: Middleware = ({
	appState,
	resource,
	resourceId,
	responseBodyMediaType: [responseBodyMediaType, responseBodySerializerPair],
	responseBodyLanguage: [languageCode, responseBodyMessageCollection],
	responseBodyEncoding: [encodingKey, encoding],
	errorResponseBodyLanguage: [errorLanguageCode, errorMessageCollection],
	errorResponseBodyMediaType: [errorMediaType, errorSerializerPair],
	errorResponseBodyEncoding: [errorEncodingKey, errorEncoding],
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const { deserializerPair: requestBodyDeserializerPair, encodingPair: requestBodyEncodingPair } = getDeserializerObjects(appState, req);
	if (typeof requestBodyDeserializerPair === 'undefined' || typeof requestBodyEncodingPair === 'undefined') {
		res.writeHead(constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToDeserializeRequest();
		res.end();
		return {
			handled: true
		};
	}

	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let existing: object | null;
	try {
		existing = await resource.dataSource.getSingle(resourceId);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToFetchResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	if (!existing) {
		res.writeHead(constants.HTTP_STATUS_NOT_FOUND, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.patchNonExistingResource(resource);
		res.end();
		return {
			handled: true
		};
	}

	let bodyDeserialized: unknown;
	try {
		const schema = resource.schema.type === 'object' ? resource.schema as v.ObjectSchema<any> : resource.schema
		bodyDeserialized = await getBody(
			req,
			requestBodyDeserializerPair,
			requestBodyEncodingPair,
			schema.type === 'object'
				? v.partial(
					schema as v.ObjectSchema<any>,
					(schema as v.ObjectSchema<any>).rest,
					(schema as v.ObjectSchema<any>).pipe
				)
				: schema
		);
	} catch (errRaw) {
		const err = errRaw as v.ValiError;
		const headers: Record<string, string> = {
			'Content-Language': languageCode,
		};
		if (!Array.isArray(err.issues)) {
			res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, headers)
			res.statusMessage = errorMessageCollection.statusMessages.invalidResourcePatch(resource);
			res.end();
			return {
				handled: true,
			};
		}
		// TODO better error reporting, localizable messages
		// TODO handle error handlers' errors
		const serialized = errorSerializerPair.serialize(
			err.issues.map((i) => (
				`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
			)),
		);
		const theFormatted = errorEncoding.encode(serialized);
		res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, {
			...headers,
			'Content-Type': errorMediaType,
			'Content-Encoding': errorEncodingKey,
		})
		res.statusMessage = errorMessageCollection.statusMessages.invalidResourcePatch(resource);
		res.end(theFormatted);
		return {
			handled: true,
		};
	}

	const params = bodyDeserialized as Record<string, unknown>;

	let newObject: object | null;
	try {
		newObject = await resource.dataSource.patch(resourceId, params);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToPatchResource(resource);
		res.end();
		return {
			handled: true,
		};
	}

	let serialized;
	try {
		serialized = responseBodySerializerPair.serialize(newObject);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToSerializeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	let theFormatted;
	try {
		theFormatted = encoding.encode(serialized);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToEncodeResponse();
		res.end();
		return {
			handled: true,
		};
	}

	res.writeHead(constants.HTTP_STATUS_OK, {
		'Content-Type': responseBodyMediaType,
		'Content-Language': languageCode,
		'Content-Encoding': encodingKey,
	});
	res.statusMessage = responseBodyMessageCollection.statusMessages.resourcePatched(resource);
	res.end(theFormatted);
	return {
		handled: true
	};

	// TODO finish the rest of the handlers!!!
};

export const handleCreateItem: Middleware = ({
	appState,
	serverParams,
	backendState,
	responseBodyMediaType: [responseBodyMediaType, responseBodySerializerPair],
	responseBodyLanguage: [languageCode, responseBodyMessageCollection],
	responseBodyEncoding: [encodingKey, encoding],
	errorResponseBodyLanguage: [errorLanguageCode, errorMessageCollection],
	errorResponseBodyMediaType: [errorMediaType, errorSerializerPair],
	errorResponseBodyEncoding: [errorEncodingKey, errorEncoding],
	resource,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const { deserializerPair: requestBodyDeserializerPair, encodingPair: requestBodyEncodingPair } = getDeserializerObjects(appState, req);
	if (typeof requestBodyDeserializerPair === 'undefined' || typeof requestBodyEncodingPair === 'undefined') {
		res.writeHead(constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToDeserializeRequest();
		res.end();
		return {
			handled: true
		};
	}

	let bodyDeserialized: unknown;
	try {
		bodyDeserialized = await getBody(
			req,
			requestBodyDeserializerPair,
			requestBodyEncodingPair,
			resource.schema
		);
	} catch (errRaw) {
		const err = errRaw as v.ValiError;
		const headers: Record<string, string> = {
			'Content-Language': errorLanguageCode,
		};
		if (!Array.isArray(err.issues)) {
			res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, headers)
			res.statusMessage = errorMessageCollection.statusMessages.invalidResource(resource);
			res.end();
			return {
				handled: true,
			};
		}
		// TODO better error reporting, localizable messages
		// TODO handle error handlers' errors
		const serialized = errorSerializerPair.serialize(
			err.issues.map((i) => (
				`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
			)),
		);
		const theFormatted = errorEncoding.encode(serialized);
		res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, {
			...headers,
			'Content-Type': errorMediaType,
			'Content-Encoding': errorEncodingKey,
		})
		res.statusMessage = errorMessageCollection.statusMessages.invalidResource(resource);
		res.end(theFormatted);
		return {
			handled: true,
		};
	}

	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	try {
		// TODO error handling for each process
		const newId = await resource.newId(resource.dataSource);
		const params = bodyDeserialized as Record<string, unknown>;
		params[resource.state.idAttr] = newId;
		const newObject = await resource.dataSource.create(params);
		let totalItemCount: number | undefined;
		if (backendState.showTotalItemCountOnCreateItem && typeof resource.dataSource.getTotalCount === 'function') {
			totalItemCount = await resource.dataSource.getTotalCount();
		}
		const headers: Record<string, string> = {
			'Content-Type': responseBodyMediaType,
			'Content-Language': languageCode,
			'Content-Encoding': encodingKey,
			'Location': `${serverParams.baseUrl}/${resource.state.routeName}/${newId}`
		};
		if (typeof totalItemCount !== 'undefined') {
			headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
		}
		const serialized = responseBodySerializerPair.serialize(newObject);
		const theFormatted = encoding.encode(serialized);
		res.writeHead(constants.HTTP_STATUS_CREATED, headers);
		res.statusMessage = responseBodyMessageCollection.statusMessages.resourceCreated(resource);
		res.end(theFormatted);
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		})
		res.statusMessage = `Could Not Return ${resource.state.itemName}`;
		res.end();
	}
	return {
		handled: true
	};
}

export const handleEmplaceItem: Middleware = ({
	appState,
	serverParams,
	responseBodyMediaType: [responseBodyMediaType, responseBodySerializerPair],
	responseBodyLanguage: [languageCode, responseBodyMessageCollection],
	responseBodyEncoding: [encodingKey, encoding],
	errorResponseBodyLanguage: [errorLanguageCode, errorMessageCollection],
	errorResponseBodyMediaType: [errorMediaType, errorSerializerPair],
	errorResponseBodyEncoding: [errorEncodingKey, errorEncoding],
	resource,
	resourceId,
	backendState,
}) => async (req: IncomingMessage, res: ServerResponse) => {
	const { deserializerPair: requestBodyDeserializerPair, encodingPair: requestBodyEncodingPair } = getDeserializerObjects(appState, req);
	if (typeof requestBodyDeserializerPair === 'undefined' || typeof requestBodyEncodingPair === 'undefined') {
		res.writeHead(constants.HTTP_STATUS_UNSUPPORTED_MEDIA_TYPE, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToDeserializeRequest();
		res.end();
		return {
			handled: true
		};
	}

	let bodyDeserialized: unknown;
	try {
		const schema = resource.schema.type === 'object' ? resource.schema as v.ObjectSchema<any> : resource.schema
		bodyDeserialized = await getBody(
			req,
			requestBodyDeserializerPair,
			requestBodyEncodingPair,
			schema.type === 'object'
				? v.merge([
					schema as v.ObjectSchema<any>,
					v.object({
						[resource.state.idAttr]: v.transform(
							v.any(),
							input => resource.state.idConfig.serialize(input),
							v.literal(resourceId)
						)
					})
				])
				: schema
		);
	} catch (errRaw) {
		const err = errRaw as v.ValiError;
		const headers: Record<string, string> = {
			'Content-Language': errorLanguageCode,
		};
		if (!Array.isArray(err.issues)) {
			res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, headers)
			res.statusMessage = errorMessageCollection.statusMessages.invalidResource(resource);
			res.end();
			return {
				handled: true,
			};
		}
		// TODO better error reporting, localizable messages
		// TODO handle error handlers' errors
		const serialized = errorSerializerPair.serialize(
			err.issues.map((i) => (
				`${i.path?.map((p) => p.key)?.join('.') ?? i.reason}: ${i.message}`
			)),
		);
		const theFormatted = errorEncoding.encode(serialized);
		res.writeHead(constants.HTTP_STATUS_BAD_REQUEST, {
			...headers,
			'Content-Type': errorMediaType,
			'Content-Encoding': errorEncodingKey,
		})
		res.statusMessage = errorMessageCollection.statusMessages.invalidResource(resource);
		res.end(theFormatted);
		return {
			handled: true,
		};
	}

	try {
		await resource.dataSource.initialize();
	} catch {
		res.writeHead(constants.HTTP_STATUS_INTERNAL_SERVER_ERROR, {
			'Content-Language': errorLanguageCode,
		});
		res.statusMessage = errorMessageCollection.statusMessages.unableToInitializeResourceDataSource(resource);
		res.end();
		return {
			handled: true
		};
	}

	try {
		// TODO error handling for each process
		const params = bodyDeserialized as Record<string, unknown>;
		params[resource.state.idAttr] = resource.state.idConfig.deserialize(params[resource.state.idAttr] as string);
		const [newObject, isCreated] = await resource.dataSource.emplace(resourceId, params);
		const serialized = responseBodySerializerPair.serialize(newObject);
		const theFormatted = encoding.encode(serialized);
		const headers: Record<string, string> = {
			'Content-Type': responseBodyMediaType,
			'Content-Language': languageCode,
			'Content-Encoding': encodingKey,
		};
		let totalItemCount: number | undefined;
		if (backendState.showTotalItemCountOnCreateItem && typeof resource.dataSource.getTotalCount === 'function') {
			totalItemCount = await resource.dataSource.getTotalCount();
		}
		if (isCreated) {
			headers['Location'] = `${serverParams.baseUrl}/${resource.state.routeName}/${resourceId}`;
			if (typeof totalItemCount !== 'undefined') {
				headers['X-Resource-Total-Item-Count'] = totalItemCount.toString();
			}
		}
		res.writeHead(isCreated ? constants.HTTP_STATUS_CREATED : constants.HTTP_STATUS_OK, headers);
		res.statusMessage = (
			isCreated
				? responseBodyMessageCollection.statusMessages.resourceCreated(resource)
				: responseBodyMessageCollection.statusMessages.resourceReplaced(resource)
		);
		res.end(theFormatted);
	} catch {
		res.statusCode = constants.HTTP_STATUS_INTERNAL_SERVER_ERROR;
		res.statusMessage = `Could Not Return ${resource.state.itemName}`;
		res.end();
	}
	return {
		handled: true
	};
}
